@@ -159,6 +159,9 @@ define service {
 """
         expected = [
             '# check myservice\n',
+            '# The following header was added automatically by juju\n',
+            '# Modifying it will affect nagios monitoring and alerting\n',
+            '# servicegroups: a\n',
             'command[check_myservice]=/usr/lib/nagios/plugins/check_http http://localhost\n',
             service_file_contents,
         ]
@@ -194,7 +197,9 @@ class NRPECheckTestCase(NRPEBaseTestCase):
     def test_valid_shortname(self):
         cases = [
             '1_number_is_fine',
+            'dots.are.good',
             'dashes-ok',
+            'UPPER_case_allowed',
             '5',
         ]
         for shortname in cases:
@@ -337,18 +342,22 @@ class NRPEMiscTestCase(NRPEBaseTestCase):
         nrpe.add_init_service_checks(bill, services, 'testunit')
         mock_isdir.assert_called_with('/var/lib/nagios')
         self.patched['call'].assert_called_with(
-            ['/usr/local/lib/nagios/plugins/check_exit_status.pl', '-s',
+            ['/usr/local/lib/nagios/plugins/check_exit_status.pl', '-e', '-s',
              '/etc/init.d/haproxy', 'status'], stdout=f,
             stderr=subprocess.STDOUT)
 
-        # Test with systemd
+        # Test regular services and snap services with systemd
+        services = ['apache2', 'haproxy', 'snap.test.test']
         self.patched['init_is_systemd'].return_value = True
         nrpe.add_init_service_checks(bill, services, 'testunit')
         expect_cmds = {
             'apache2': '/usr/lib/nagios/plugins/check_systemd.py apache2',
             'haproxy': '/usr/lib/nagios/plugins/check_systemd.py haproxy',
+            'snap.test.test': '/usr/lib/nagios/plugins/check_systemd.py snap.test.test',
         }
         self.assertEqual(bill.checks[2].shortname, 'apache2')
         self.assertEqual(bill.checks[2].check_cmd, expect_cmds['apache2'])
         self.assertEqual(bill.checks[3].shortname, 'haproxy')
         self.assertEqual(bill.checks[3].check_cmd, expect_cmds['haproxy'])
+        self.assertEqual(bill.checks[4].shortname, 'snap.test.test')
+        self.assertEqual(bill.checks[4].check_cmd, expect_cmds['snap.test.test'])
