@@ -49,9 +49,12 @@ DUMMY_ADDRESSES = {
     'eth1': {
         2: [{'addr': '10.5.0.1',
              'broadcast': '10.5.255.255',
-             'netmask': '255.255.0.0'}],
-        10: [{'addr': 'fe80::3e97:eff:fe8b:1cf7%eth1',
-              'netmask': 'ffff:ffff:ffff:ffff::'}],
+             'netmask': '255.255.0.0'},
+            {'addr': '10.6.0.2',
+             'broadcast': '10.6.0.255',
+             'netmask': '255.255.255.0'}],
+        3: [{'addr': 'fe80::3e97:eff:fe8b:1cf7%eth1',
+             'netmask': 'ffff:ffff:ffff:ffff::'}],
         17: [{'addr': '3c:97:0e:8b:1c:f7',
               'broadcast': 'ff:ff:ff:ff:ff:ff'}]
     },
@@ -197,6 +200,10 @@ class IPTest(unittest.TestCase):
         self._test_get_address_in_network('192.168.10.58',
                                           '192.168.11.0/24 192.168.10.0/24')
 
+    def test_get_address_in_network_ipv4_secondary(self):
+        self._test_get_address_in_network('10.6.0.2',
+                                          '10.6.0.0/24')
+
     def test_get_address_in_network_ipv6(self):
         self._test_get_address_in_network('2a01:348:2f4:0:685e:5748:ae62:209f',
                                           '2a01:348:2f4::/64')
@@ -285,7 +292,8 @@ class IPTest(unittest.TestCase):
             '255.255.0.0')
         self.assertEquals(net_ip.get_netmask_for_address('172.4.5.5'), None)
         self.assertEquals(
-            net_ip.get_netmask_for_address('2a01:348:2f4:0:685e:5748:ae62:210f'),
+            net_ip.get_netmask_for_address(
+                '2a01:348:2f4:0:685e:5748:ae62:210f'),
             '64'
         )
         self.assertEquals(
@@ -372,6 +380,28 @@ class IPTest(unittest.TestCase):
         _interfaces.return_value = DUMMY_ADDRESSES.keys()
         self.assertRaises(Exception, net_ip.get_ipv6_addr, 'eth1')
 
+    @patch('charmhelpers.contrib.network.ip.subprocess.check_output')
+    def test_is_ipv6_disabled(self, mock_check_output):
+        # verify that the function does look for the right thing
+        mock_check_output.return_value = """
+        Some lines before
+        net.ipv6.conf.all.disable_ipv6 = 1
+        Some lines afterward
+        """
+        self.assertTrue(net_ip.is_ipv6_disabled())
+        mock_check_output.assert_called_once_with(
+            ['sysctl', 'net.ipv6.conf.all.disable_ipv6'],
+            stderr=subprocess.STDOUT, universal_newlines=True)
+        # if it isn't there, it must return false
+        mock_check_output.return_value = ""
+        self.assertFalse(net_ip.is_ipv6_disabled())
+        # If the syscall returns an error, then return True
+
+        def fake_check_call(*args, **kwargs):
+            raise subprocess.CalledProcessError(['called'], 1)
+        mock_check_output.side_effect = fake_check_call
+        self.assertTrue(net_ip.is_ipv6_disabled())
+
     @patch.object(netifaces, 'ifaddresses')
     @patch.object(netifaces, 'interfaces')
     def test_get_iface_addr(self, _interfaces, _ifaddresses):
@@ -554,9 +584,8 @@ class IPTest(unittest.TestCase):
 
     @patch('charmhelpers.contrib.network.ip.subprocess.check_output')
     @patch('charmhelpers.contrib.network.ip.get_iface_addr')
-    def test_get_ipv6_global_dynamic_address_invalid_address(self,
-                                                             mock_get_iface_addr,
-                                                             mock_check_out):
+    def test_get_ipv6_global_dynamic_address_invalid_address(
+            self, mock_get_iface_addr, mock_check_out):
         mock_get_iface_addr.return_value = []
         with nose.tools.assert_raises(Exception):
             net_ip.get_ipv6_addr()
@@ -774,54 +803,45 @@ class IPTest(unittest.TestCase):
     def test_get_relation_ip(self, assert_charm_supports_ipv6, get_ipv6_addr,
                              unit_get, get_address_in_network,
                              network_get_primary_address, config):
+        ACCESS_IP = '10.50.1.1'
+        ACCESS_NETWORK = '10.50.1.0/24'
         AMQP_IP = '10.200.1.1'
-        OVERRIDE_AMQP_IP = '10.250.1.1'
-        CLUSTER_IP = '10.100.1.1'
-        OVERRIDE_CLUSTER_IP = '10.150.1.1'
         IPV6_IP = '2001:DB8::1'
         DEFAULT_IP = '172.16.1.1'
         assert_charm_supports_ipv6.return_value = True
         get_ipv6_addr.return_value = [IPV6_IP]
         unit_get.return_value = DEFAULT_IP
         get_address_in_network.return_value = DEFAULT_IP
-        network_get_primary_address.return_value = DEFAULT_IP
-
-        # IPv6
-        _config = {'prefer-ipv6': True,
-                   'cluster-network': '10.100.1.0/24',
-                   'access-network': '10.200.1.0/24'}
-        config.side_effect = lambda key: _config.get(key)
-        self.assertEqual(IPV6_IP, net_ip.get_relation_ip('amqp'))
-
-        # Overrides
-        _config = {'prefer-ipv6': False,
-                   'cluster-network': '10.100.1.0/24',
-                   'access-network': '10.200.1.0/24'}
-        config.side_effect = lambda key: _config.get(key)
-
-        get_address_in_network.return_value = OVERRIDE_AMQP_IP
-        self.assertEqual(
-            OVERRIDE_AMQP_IP,
-            net_ip.get_relation_ip('amqp',
-                                   config_override='access-network'))
-
-        get_address_in_network.return_value = OVERRIDE_CLUSTER_IP
-        self.assertEqual(OVERRIDE_CLUSTER_IP,
-                         net_ip.get_relation_ip(
-                             'cluster',
-                             config_override='cluster-network'))
+        network_get_primary_address.return_value = AMQP_IP
 
         # Network-get calls
-        _config = {'prefer-ipv6': False,
-                   'cluster-network': None,
-                   'access-network': None}
+        _config = {'prefer-ipv6': False}
         config.side_effect = lambda key: _config.get(key)
 
-        network_get_primary_address.return_value = AMQP_IP
+        network_get_primary_address.side_effect = NotImplementedError
+        self.assertEqual(DEFAULT_IP, net_ip.get_relation_ip('amqp'))
+
+        network_get_primary_address.side_effect = net_ip.NoNetworkBinding
+        self.assertEqual(DEFAULT_IP, net_ip.get_relation_ip('doesnotexist'))
+
+        network_get_primary_address.side_effect = None
         self.assertEqual(AMQP_IP, net_ip.get_relation_ip('amqp'))
 
-        network_get_primary_address.return_value = CLUSTER_IP
-        self.assertEqual(CLUSTER_IP,
-                         net_ip.get_relation_ip(
-                             config_override='cluster-network',
-                             interface='cluster'))
+        self.assertFalse(get_address_in_network.called)
+
+        # Specific CIDR network
+        get_address_in_network.return_value = ACCESS_IP
+        network_get_primary_address.return_value = DEFAULT_IP
+        self.assertEqual(
+            ACCESS_IP,
+            net_ip.get_relation_ip('shared-db',
+                                   cidr_network=ACCESS_NETWORK))
+        get_address_in_network.assert_called_with(ACCESS_NETWORK, DEFAULT_IP)
+
+        self.assertFalse(assert_charm_supports_ipv6.called)
+
+        # IPv6
+        _config = {'prefer-ipv6': True}
+        config.side_effect = lambda key: _config.get(key)
+        self.assertEqual(IPV6_IP, net_ip.get_relation_ip('amqp'))
+        assert_charm_supports_ipv6.assert_called_with()
