@@ -7,8 +7,11 @@ from threading import Timer
 from testtools import TestCase
 import json
 import copy
+import shutil
 
 import charmhelpers.contrib.storage.linux.ceph as ceph_utils
+
+from charmhelpers.core.unitdata import Storage
 from subprocess import CalledProcessError
 from tests.helpers import patch_open, FakeRelation
 import nose.plugins.attrib
@@ -33,7 +36,7 @@ bar
 baz
 """
 # Vastly abbreviated output from ceph osd dump --format=json
-OSD_DUMP = """
+OSD_DUMP = b"""
 {
     "pools": [
         {
@@ -82,7 +85,7 @@ OSD_DUMP = """
 }
 """
 
-MONMAP_DUMP = """{
+MONMAP_DUMP = b"""{
     "name": "ip-172-31-13-119", "rank": 0, "state": "leader",
     "election_epoch": 18, "quorum": [0, 1, 2],
     "outside_quorum": [],
@@ -209,6 +212,16 @@ class CephUtilsTests(TestCase):
                           valid_type=six.string_types,
                           valid_range=["valid", "list", "of", "strings"])
 
+    def test_validator_valid_string(self):
+        ceph_utils.validator(value="foo",
+                             valid_type=six.string_types,
+                             valid_range=["foo"])
+
+    def test_validator_valid_string_type(self):
+        ceph_utils.validator(value="foo",
+                             valid_type=str,
+                             valid_range=["foo"])
+
     def test_pool_add_cache_tier(self):
         p = ceph_utils.Pool(name='test', service='admin')
         p.add_cache_tier('cacher', 'readonly')
@@ -391,11 +404,25 @@ class CephUtilsTests(TestCase):
         return_value = ceph_utils.get_erasure_profile('admin', 'unknown')
         self.assertEqual(None, return_value)
 
-    def test_pool_set(self):
+    def test_pool_set_int(self):
         self.check_call.return_value = 0
         ceph_utils.pool_set(service='admin', pool_name='data', key='test', value=2)
         self.check_call.assert_has_calls([
-            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'data', 'test', 2])
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'data', 'test', '2'])
+        ])
+
+    def test_pool_set_bool(self):
+        self.check_call.return_value = 0
+        ceph_utils.pool_set(service='admin', pool_name='data', key='test', value=True)
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'data', 'test', 'true'])
+        ])
+
+    def test_pool_set_str(self):
+        self.check_call.return_value = 0
+        ceph_utils.pool_set(service='admin', pool_name='data', key='test', value='two')
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'data', 'test', 'two'])
         ])
 
     def test_pool_set_fails(self):
@@ -438,15 +465,26 @@ class CephUtilsTests(TestCase):
             call(['ceph', '--id', 'admin', 'osd', 'pool', 'set-quota', 'data', 'max_bytes', '0'])
         ])
 
+    @patch.object(ceph_utils, 'ceph_version')
     @patch.object(ceph_utils, 'erasure_profile_exists')
-    def test_create_erasure_profile(self, existing_profile):
+    def test_create_erasure_profile(self, existing_profile, mock_version):
         existing_profile.return_value = True
+        mock_version.return_value = '10.0.0'
         ceph_utils.create_erasure_profile(service='admin', profile_name='super-profile', erasure_plugin_name='jerasure',
                                           failure_domain='rack', data_chunks=10, coding_chunks=3)
 
         cmd = ['ceph', '--id', 'admin', 'osd', 'erasure-code-profile', 'set', 'super-profile',
                'plugin=' + 'jerasure', 'k=' + str(10), 'm=' + str(3),
-               'ruleset_failure_domain=' + 'rack', '--force']
+               'ruleset-failure-domain=' + 'rack', '--force']
+        self.check_call.assert_has_calls([call(cmd)])
+
+        mock_version.return_value = '12.1.0'
+        ceph_utils.create_erasure_profile(service='admin', profile_name='super-profile', erasure_plugin_name='jerasure',
+                                          failure_domain='rack', data_chunks=10, coding_chunks=3)
+
+        cmd = ['ceph', '--id', 'admin', 'osd', 'erasure-code-profile', 'set', 'super-profile',
+               'plugin=' + 'jerasure', 'k=' + str(10), 'm=' + str(3),
+               'crush-failure-domain=' + 'rack', '--force']
         self.check_call.assert_has_calls([call(cmd)])
 
     @patch.object(ceph_utils, 'erasure_profile_exists')
@@ -457,7 +495,7 @@ class CephUtilsTests(TestCase):
 
         cmd = ['ceph', '--id', 'admin', 'osd', 'erasure-code-profile', 'set', 'super-profile',
                'plugin=' + 'local', 'k=' + str(10), 'm=' + str(3),
-               'ruleset_failure_domain=' + 'rack', 'l=' + str(1)]
+               'ruleset-failure-domain=' + 'rack', 'l=' + str(1)]
         self.check_call.assert_has_calls([call(cmd)])
 
     @patch.object(ceph_utils, 'erasure_profile_exists')
@@ -469,7 +507,7 @@ class CephUtilsTests(TestCase):
 
         cmd = ['ceph', '--id', 'admin', 'osd', 'erasure-code-profile', 'set', 'super-profile',
                'plugin=' + 'shec', 'k=' + str(10), 'm=' + str(3),
-               'ruleset_failure_domain=' + 'rack', 'c=' + str(1)]
+               'ruleset-failure-domain=' + 'rack', 'c=' + str(1)]
         self.check_call.assert_has_calls([call(cmd)])
 
     def test_rename_pool(self):
@@ -547,7 +585,10 @@ class CephUtilsTests(TestCase):
             '010d57d581604d411b315dd64112bff832ab92c7323fa06077134b50',
             '8e0a9705c1aeafa1ce250cc9f1bb443fc6e5150e5edcbeb6eeb82e3c',
             'c3f8d36ba098c23ee920cb08cfb9beda6b639f8433637c190bdd56ec']
-        monmap.return_value = json.loads(MONMAP_DUMP)
+        _monmap_dump = MONMAP_DUMP
+        if six.PY3:
+            _monmap_dump = _monmap_dump.decode('UTF-8')
+        monmap.return_value = json.loads(_monmap_dump)
         hashed_mon_list = ceph_utils.hash_monitor_names(service='admin')
         self.assertEqual(expected=expected_hash_list, observed=hashed_mon_list)
 
@@ -1018,6 +1059,17 @@ class CephUtilsTests(TestCase):
         rget.return_value = ''
         self.assertEquals(False, ceph_utils.ensure_ceph_keyring(service='foo'))
 
+    @patch.object(ceph_utils, '_keyring_path')
+    @patch.object(ceph_utils, 'create_keyring')
+    @patch.object(ceph_utils, 'relation_ids')
+    def test_ensure_ceph_keyring_no_relation_but_key(self, rids,
+                                                     create, _path):
+        rids.return_value = []
+        self.assertTrue(ceph_utils.ensure_ceph_keyring(service='foo',
+                                                       key='testkey'))
+        create.assert_called_with(service='foo', key='testkey')
+        _path.assert_called_with('foo')
+
     @patch.object(ceph_utils, '_keyring_path')
     @patch.object(ceph_utils, 'create_keyring')
     @patch.object(ceph_utils, 'relation_ids')
@@ -1079,6 +1131,9 @@ class CephUtilsTests(TestCase):
         rq.add_op_request_access_to_group(name='test')
         rq.add_op_request_access_to_group(name='objects',
                                           key_name='test')
+        rq.add_op_request_access_to_group(
+            name='others',
+            object_prefix_permissions={'rwx': ['prefix1']})
         expected = {
             'api-version': 1,
             'request-id': 'uuid',
@@ -1086,14 +1141,38 @@ class CephUtilsTests(TestCase):
                     {'op': 'create-pool', 'name': 'pool2', 'replicas': 3},
                     {'op': 'create-pool', 'name': 'pool3', 'replicas': 3, 'group': 'test'},
                     {'op': 'add-permissions-to-key', 'group': 'test', 'name': 'service_test'},
-                    {'op': 'add-permissions-to-key', 'group': 'objects', 'name': 'test'}]
+                    {'op': 'add-permissions-to-key', 'group': 'objects', 'name': 'test'},
+                    {
+                        'op': 'add-permissions-to-key',
+                        'group': 'others',
+                        'name': 'service_test',
+                        'object-prefix-permissions': {u'rwx': [u'prefix1']}}]
         }
         request_dict = json.loads(rq.request)
         for key in ['api-version', 'request-id']:
             self.assertEqual(request_dict[key], expected[key])
-        for key in ['op', 'name', 'replicas']:
-            self.assertEqual(request_dict['ops'][0][key], expected['ops'][0][key])
-            self.assertEqual(request_dict['ops'][1][key], expected['ops'][1][key])
+        for (op_no, expected_op) in enumerate(expected['ops']):
+            for key in expected_op.keys():
+                self.assertEqual(
+                    request_dict['ops'][op_no][key],
+                    expected_op[key])
+
+    @patch.object(ceph_utils, 'service_name')
+    @patch.object(ceph_utils, 'uuid')
+    def test_ceph_broker_rq_class_test_not_equal(self, uuid, service_name):
+        service_name.return_value = 'service_test'
+        uuid.uuid1.return_value = 'uuid'
+        rq1 = ceph_utils.CephBrokerRq()
+        rq1.add_op_create_pool('pool1')
+        rq1.add_op_request_access_to_group(name='test')
+        rq1.add_op_request_access_to_group(name='objects',
+                                           permission='rwx')
+        rq2 = ceph_utils.CephBrokerRq()
+        rq2.add_op_create_pool('pool1')
+        rq2.add_op_request_access_to_group(name='test')
+        rq2.add_op_request_access_to_group(name='objects',
+                                           permission='r')
+        self.assertFalse(rq1 == rq2)
 
     def test_ceph_broker_rsp_class(self):
         rsp = ceph_utils.CephBrokerRsp(json.dumps({'exit-code': 0,
@@ -1386,3 +1465,52 @@ class CephUtilsTests(TestCase):
         mock_config.return_value = ("{'osd': {'foo': 1},"
                                     "'unknown': {'blah': 1}}")
         self.assertEqual({'osd': {'foo': 1}}, ctxt)
+
+    @patch.object(ceph_utils, 'local_unit', lambda: "nova-compute/0")
+    def test_is_broker_action_done(self):
+        tmpdir = mkdtemp()
+        try:
+            db_path = '{}/kv.db'.format(tmpdir)
+            with patch('charmhelpers.core.unitdata._KV', Storage(db_path)):
+                rq_id = "3d03e9f6-4c36-11e7-89ba-fa163e7c7ec6"
+                broker_key = ceph_utils.get_broker_rsp_key()
+                self.relation_get.return_value = {broker_key:
+                                                  json.dumps({"request-id":
+                                                              rq_id,
+                                                              "exit-code": 0})}
+                action = 'restart_nova_compute'
+                ret = ceph_utils.is_broker_action_done(action, rid="ceph:1",
+                                                       unit="ceph/0")
+                self.relation_get.assert_has_calls([call('ceph:1', 'ceph/0')])
+                self.assertFalse(ret)
+
+                ceph_utils.mark_broker_action_done(action)
+                self.assertTrue(os.path.exists(tmpdir))
+                ret = ceph_utils.is_broker_action_done(action, rid="ceph:1",
+                                                       unit="ceph/0")
+                self.assertTrue(ret)
+        finally:
+            if os.path.exists(tmpdir):
+                shutil.rmtree(tmpdir)
+
+    @patch.object(ceph_utils, 'local_unit', lambda: "nova-compute/0")
+    def test_mark_broker_action_done(self):
+        tmpdir = mkdtemp()
+        try:
+            db_path = '{}/kv.db'.format(tmpdir)
+            with patch('charmhelpers.core.unitdata._KV', Storage(db_path)):
+                rq_id = "3d03e9f6-4c36-11e7-89ba-fa163e7c7ec6"
+                broker_key = ceph_utils.get_broker_rsp_key()
+                self.relation_get.return_value = {broker_key:
+                                                  json.dumps({"request-id":
+                                                              rq_id})}
+                action = 'restart_nova_compute'
+                ceph_utils.mark_broker_action_done(action, rid="ceph:1",
+                                                   unit="ceph/0")
+                key = 'unit_0_ceph_broker_action.{}'.format(action)
+                self.relation_get.assert_has_calls([call('ceph:1', 'ceph/0')])
+                kvstore = Storage(db_path)
+                self.assertEqual(kvstore.get(key=key), rq_id)
+        finally:
+            if os.path.exists(tmpdir):
+                shutil.rmtree(tmpdir)
