@@ -116,6 +116,7 @@ class ConfigTest(TestCase):
         with open(c.path, 'r') as f:
             self.assertEqual(c, json.load(f))
             self.assertEqual(c, dict(foo='bar'))
+            self.assertEqual(os.stat(c.path).st_mode & 0o777, 0o600)
 
     def test_save_with_prev_dict(self):
         c = hookenv.Config(dict(foo='bar'))
@@ -126,17 +127,18 @@ class ConfigTest(TestCase):
         with open(c.path, 'r') as f:
             self.assertEqual(c, json.load(f))
             self.assertEqual(c, dict(foo='bar', a='b'))
+            self.assertEqual(os.stat(c.path).st_mode & 0o777, 0o600)
 
     def test_deep_change(self):
         # After loading stored data into our previous dictionary,
         # it gets copied into our current dictionary. If this is not
         # a deep copy, then mappings and lists will be shared instances
         # and changes will not be detected.
-        c = hookenv.Config(dict(l=[]))
+        c = hookenv.Config(dict(ll=[]))
         c.save()
         c = hookenv.Config()
-        c['l'].append(42)
-        self.assertTrue(c.changed('l'), 'load_previous() did not deepcopy')
+        c['ll'].append(42)
+        self.assertTrue(c.changed('ll'), 'load_previous() did not deepcopy')
 
     def test_getitem(self):
         c = hookenv.Config(dict(foo='bar'))
@@ -1105,10 +1107,12 @@ class HelpersTest(TestCase):
         hookenv.open_port(443, "TCP")
         hookenv.open_port(80)
         hookenv.open_port(100, "UDP")
+        hookenv.open_port(0, "ICMP")
         calls = [
             call(['open-port', '443/TCP']),
             call(['open-port', '80/TCP']),
             call(['open-port', '100/UDP']),
+            call(['open-port', 'ICMP']),
         ]
         check_call_.assert_has_calls(calls)
 
@@ -1117,10 +1121,12 @@ class HelpersTest(TestCase):
         hookenv.close_port(443, "TCP")
         hookenv.close_port(80)
         hookenv.close_port(100, "UDP")
+        hookenv.close_port(0, "ICMP")
         calls = [
             call(['close-port', '443/TCP']),
             call(['close-port', '80/TCP']),
             call(['close-port', '100/UDP']),
+            call(['close-port', 'ICMP']),
         ]
         check_call_.assert_has_calls(calls)
 
@@ -1148,6 +1154,13 @@ class HelpersTest(TestCase):
         ]
         check_call_.assert_has_calls(calls)
 
+    @patch('subprocess.check_output')
+    def test_gets_opened_ports(self, check_output_):
+        prts = ['8080/tcp', '8081-8083/tcp']
+        check_output_.return_value = json.dumps(prts).encode('UTF-8')
+        self.assertEqual(hookenv.opened_ports(), prts)
+        check_output_.assert_called_with(['opened-ports', '--format=json'])
+
     @patch('subprocess.check_output')
     def test_gets_unit_attribute(self, check_output_):
         check_output_.return_value = json.dumps('bar').encode('UTF-8')
@@ -1161,11 +1174,17 @@ class HelpersTest(TestCase):
         check_output_.assert_called_with(['unit-get', '--format=json', 'foo'])
 
     def test_cached_decorator(self):
+        class Unserializable(object):
+            def __str__(self):
+                return 'unserializable'
+
+        unserializable = Unserializable()
         calls = []
         values = {
             'hello': 'world',
             'foo': 'bar',
             'baz': None,
+            unserializable: 'qux',
         }
 
         @hookenv.cached
@@ -1178,11 +1197,16 @@ class HelpersTest(TestCase):
         self.assertEquals(cache_function('foo'), 'bar')
         self.assertEquals(cache_function('baz'), None)
         self.assertEquals(cache_function('baz'), None)
-        self.assertEquals(calls, ['hello', 'foo', 'baz'])
+        self.assertEquals(cache_function(unserializable), 'qux')
+        self.assertEquals(calls, ['hello', 'foo', 'baz', unserializable])
 
     def test_gets_charm_dir(self):
+        with patch.dict('os.environ', {}):
+            self.assertEqual(hookenv.charm_dir(), None)
         with patch.dict('os.environ', {'CHARM_DIR': '/var/empty'}):
             self.assertEqual(hookenv.charm_dir(), '/var/empty')
+        with patch.dict('os.environ', {'JUJU_CHARM_DIR': '/var/another'}):
+            self.assertEqual(hookenv.charm_dir(), '/var/another')
 
     @patch('subprocess.check_output')
     def test_resource_get_unsupported(self, check_output_):
@@ -1467,9 +1491,10 @@ class HooksTest(TestCase):
 
         juju_version.return_value = '1.24-beta5.1-trusty-amd64'
         self.assertTrue(hookenv.has_juju_version('1.23'))
-        self.assertFalse(hookenv.has_juju_version('1.24'))
+        self.assertTrue(hookenv.has_juju_version('1.24'))  # Better if this was false!
         self.assertTrue(hookenv.has_juju_version('1.24-beta5'))
         self.assertTrue(hookenv.has_juju_version('1.24-beta5.1'))
+        self.assertFalse(hookenv.has_juju_version('1.25'))
         self.assertTrue(hookenv.has_juju_version('1.18-backport6'))
 
     @patch.object(hookenv, 'relation_to_role_and_interface')
@@ -1650,20 +1675,96 @@ class HooksTest(TestCase):
 
     @patch('subprocess.check_output')
     def test_network_get_primary(self, check_output):
-        '''Ensure that network-get is called correctly and output is returned'''
+        """Ensure that network-get is called correctly and output is returned"""
         check_output.return_value = b'192.168.22.1'
         ip = hookenv.network_get_primary_address('mybinding')
         check_output.assert_called_with(
-            ['network-get', '--primary-address', 'mybinding'])
+            ['network-get', '--primary-address', 'mybinding'], stderr=-2)
         self.assertEqual(ip, '192.168.22.1')
 
     @patch('subprocess.check_output')
     def test_network_get_primary_unsupported(self, check_output):
-        '''Ensure that NotImplementedError is thrown when run on Juju < 2.0'''
+        """Ensure that NotImplementedError is thrown when run on Juju < 2.0"""
         check_output.side_effect = OSError(2, 'network-get')
         self.assertRaises(NotImplementedError, hookenv.network_get_primary_address,
                           'mybinding')
 
+    @patch('subprocess.check_output')
+    def test_network_get_primary_no_binding_found(self, check_output):
+        """Ensure that NotImplementedError when no binding is found"""
+        check_output.side_effect = CalledProcessError(
+            1, 'network-get',
+            output='no network config found for binding'.encode('UTF-8'))
+        self.assertRaises(hookenv.NoNetworkBinding,
+                          hookenv.network_get_primary_address,
+                          'doesnotexist')
+        check_output.assert_called_with(
+            ['network-get', '--primary-address', 'doesnotexist'], stderr=-2)
+
+    @patch('subprocess.check_output')
+    def test_network_get_primary_other_exception(self, check_output):
+        """Ensure that CalledProcessError still thrown when not
+        a missing binding"""
+        check_output.side_effect = CalledProcessError(
+            1, 'network-get',
+            output='any other message'.encode('UTF-8'))
+        self.assertRaises(CalledProcessError,
+                          hookenv.network_get_primary_address,
+                          'mybinding')
+
+    @patch('charmhelpers.core.hookenv.juju_version')
+    @patch('subprocess.check_output')
+    def test_network_get(self, check_output, juju_version):
+        """Ensure that network-get is called correctly"""
+        juju_version.return_value = '2.2.0'
+        check_output.return_value = b'result'
+        hookenv.network_get('endpoint')
+        check_output.assert_called_with(
+            ['network-get', 'endpoint', '--format', 'yaml'], stderr=-2)
+
+    @patch('charmhelpers.core.hookenv.juju_version')
+    @patch('subprocess.check_output')
+    def test_network_get_primary_required(self, check_output, juju_version):
+        """Ensure that NotImplementedError is thrown with Juju < 2.2.0"""
+        check_output.return_value = b'result'
+
+        juju_version.return_value = '2.1.4'
+        self.assertRaises(NotImplementedError, hookenv.network_get, 'binding')
+        juju_version.return_value = '2.2.0'
+        self.assertEquals(hookenv.network_get('endpoint'), 'result')
+
+    @patch('charmhelpers.core.hookenv.juju_version')
+    @patch('subprocess.check_output')
+    def test_network_get_relation_bound(self, check_output, juju_version):
+        """Ensure that network-get supports relation context, requires Juju 2.3"""
+        juju_version.return_value = '2.3.0'
+        check_output.return_value = b'result'
+        hookenv.network_get('endpoint', 'db')
+        check_output.assert_called_with(
+            ['network-get', 'endpoint', '--format', 'yaml', '-r', 'db'],
+            stderr=-2)
+        juju_version.return_value = '2.2.8'
+        self.assertRaises(NotImplementedError, hookenv.network_get, 'endpoint', 'db')
+
+    @patch('charmhelpers.core.hookenv.juju_version')
+    @patch('subprocess.check_output')
+    def test_network_get_parses_yaml(self, check_output, juju_version):
+        """network-get returns loaded YAML output."""
+        juju_version.return_value = '2.3.0'
+        check_output.return_value = b"""
+bind-addresses:
+- macaddress: ""
+  interfacename: ""
+  addresses:
+    - address: 10.136.107.33
+      cidr: ""
+ingress-addresses:
+- 10.136.107.33
+        """
+        ip = hookenv.network_get('mybinding')
+        self.assertEqual(len(ip['bind-addresses']), 1)
+        self.assertEqual(ip['ingress-addresses'], ['10.136.107.33'])
+
     @patch('subprocess.check_call')
     def test_add_metric(self, check_call_):
         hookenv.add_metric(flips='1.5', flops='2.1')
@@ -1691,3 +1792,60 @@ class HooksTest(TestCase):
         }
         self.assertEqual(hookenv.meter_status(), 'GREEN')
         self.assertEqual(hookenv.meter_info(), 'all good')
+
+    @patch.object(hookenv, 'related_units')
+    @patch.object(hookenv, 'relation_ids')
+    def test_iter_units_for_relation_name(self, relation_ids, related_units):
+        relation_ids.return_value = ['rel:1']
+        related_units.return_value = ['unit/0', 'unit/1', 'unit/2']
+        expected = [('rel:1', 'unit/0'),
+                    ('rel:1', 'unit/1'),
+                    ('rel:1', 'unit/2')]
+        related_units_data = [
+            (u.rid, u.unit)
+            for u in hookenv.iter_units_for_relation_name('rel')]
+        self.assertEqual(expected, related_units_data)
+
+    @patch.object(hookenv, 'relation_get')
+    def test_ingress_address(self, relation_get):
+        """Ensure ingress_address returns the ingress-address when available
+        and returns the private-address when not.
+        """
+        _with_ingress = {'egress-subnets': '10.5.0.23/32',
+                         'ingress-address': '10.5.0.23',
+                         'private-address': '172.16.5.10'}
+
+        _without_ingress = {'private-address': '172.16.5.10'}
+
+        # Return the ingress-address
+        relation_get.return_value = _with_ingress
+        self.assertEqual(hookenv.ingress_address(rid='test:1', unit='unit/1'),
+                         '10.5.0.23')
+        relation_get.assert_called_with(rid='test:1', unit='unit/1')
+        # Return the private-address
+        relation_get.return_value = _without_ingress
+        self.assertEqual(hookenv.ingress_address(rid='test:1'),
+                         '172.16.5.10')
+
+    @patch.object(hookenv, 'relation_get')
+    def test_egress_subnets(self, relation_get):
+        """Ensure egress_subnets returns the decoded egress-subnets when available
+        and falls back correctly when not.
+        """
+        d = {'egress-subnets': '10.5.0.23/32,2001::F00F/64',
+             'ingress-address': '10.5.0.23',
+             'private-address': '2001::D0:F00D'}
+
+        # Return the egress-subnets
+        relation_get.return_value = d
+        self.assertEqual(hookenv.egress_subnets(rid='test:1', unit='unit/1'),
+                         ['10.5.0.23/32', '2001::F00F/64'])
+        relation_get.assert_called_with(rid='test:1', unit='unit/1')
+
+        # Return the ingress-address
+        del d['egress-subnets']
+        self.assertEqual(hookenv.egress_subnets(), ['10.5.0.23/32'])
+
+        # Return the private-address
+        del d['ingress-address']
+        self.assertEqual(hookenv.egress_subnets(), ['2001::D0:F00D/128'])
