@@ -193,7 +193,8 @@ class HelpersTest(TestCase):
         self.assertTrue(host.service_pause(service_name))
         service.assert_has_calls([
             call('stop', service_name),
-            call('disable', service_name)])
+            call('disable', service_name),
+            call('mask', service_name)])
 
     @patch.object(host, 'service_running')
     @patch.object(host, 'init_is_systemd')
@@ -206,6 +207,9 @@ class HelpersTest(TestCase):
         systemd.return_value = True
         self.assertTrue(host.service_resume(service_name))
         service.assert_has_calls([
+            call('unmask', service_name),
+            # Ensures a package starts up if disabled but not masked,
+            # per lp:1692178
             call('enable', service_name),
             call('start', service_name)])
 
@@ -928,6 +932,40 @@ class HelpersTest(TestCase):
         ])
         getgrnam.assert_called_with(group_name)
 
+    @patch('subprocess.check_call')
+    def test_chage_no_chroot(self, check_call):
+        host.chage('usera', expiredate='2019-09-28', maxdays='11')
+        check_call.assert_called_with([
+            'chage',
+            '--expiredate', '2019-09-28',
+            '--maxdays', '11',
+            'usera'
+        ])
+
+    @patch('subprocess.check_call')
+    def test_chage_chroot(self, check_call):
+        host.chage('usera', expiredate='2019-09-28', maxdays='11',
+                   root='mychroot')
+        check_call.assert_called_with([
+            'chage',
+            '--root', 'mychroot',
+            '--expiredate', '2019-09-28',
+            '--maxdays', '11',
+            'usera'
+        ])
+
+    @patch('subprocess.check_call')
+    def test_remove_password_expiry(self, check_call):
+        host.remove_password_expiry('usera')
+        check_call.assert_called_with([
+            'chage',
+            '--expiredate', '-1',
+            '--inactive', '-1',
+            '--mindays', '0',
+            '--maxdays', '-1',
+            'usera'
+        ])
+
     @patch('subprocess.check_output')
     @patch.object(host, 'log')
     def test_rsyncs_a_path(self, log, check_output):
@@ -1814,4 +1852,18 @@ class HelpersTest(TestCase):
 
         self.assertTrue(handle.read.call_count == 1)
         self.assertTrue(handle.seek.call_count == 1)
-        handle.write.assert_called_once_with('PRUNEPATHS="/tmp /srv/node /tmp/test"')
+        handle.write.assert_called_once_with(
+            'PRUNEPATHS="/tmp /srv/node /tmp/test"')
+
+    @patch.object(host, 'local_unit')
+    def test_modulo_distribution(self, local_unit):
+        local_unit.return_value = 'test/7'
+        self.assertEqual(host.modulo_distribution(modulo=6, wait=10), 10)
+
+
+class TestHostCompator(TestCase):
+
+    def test_compare_ubuntu_releases(self):
+        from charmhelpers.osplatform import get_platform
+        if get_platform() == 'ubuntu':
+            self.assertTrue(host.CompareHostReleases('yakkety') < 'zesty')
